from __future__ import annotations

import contextlib
import datetime
import typing
from typing import Literal

import numpy as np
import pandas as pd
import pytz

from exchange_calendars import errors

if typing.TYPE_CHECKING:
    from exchange_calendars import ExchangeCalendar

NANOSECONDS_PER_MINUTE = int(6e10)

NP_NAT = pd.NaT.value

# Use Date type where input does not need to represent an actual session
# and will be parsed by parse_date.
Date = typing.Union[pd.Timestamp, str, int, float, datetime.datetime]

# Use Session type where input should represent an actual session and will
# be parsed by parse_session.
Session = Date

# Use Minute type where input does not need to represent an actual trading
# minute and will be parsed by parse_timestamp.
Minute = typing.Union[pd.Timestamp, str, int, float, datetime.datetime]

# Use TradingMinute where input should represent a trading minute and will
# be parsed by parse_trading_minute.
TradingMinute = Minute


def next_divider_idx(dividers: np.ndarray, minute_val: int) -> int:

    divider_idx = np.searchsorted(dividers, minute_val, side="right")
    target = dividers[divider_idx]

    if minute_val == target:
        # if dt is exactly on the divider, go to the next value
        return divider_idx + 1
    else:
        return divider_idx


def previous_divider_idx(dividers: np.ndarray, minute_val: int) -> int:

    divider_idx = np.searchsorted(dividers, minute_val)

    if divider_idx == 0:
        raise ValueError("Cannot go earlier in calendar!")

    return divider_idx - 1


def compute_minutes(
    opens_in_ns: np.ndarray,
    break_starts_in_ns: np.ndarray,
    break_ends_in_ns: np.ndarray,
    closes_in_ns: np.ndarray,
    side: Literal["left", "right", "both", "neither"] = "both",
) -> np.ndarray:
    """Return array of trading minutes."""
    start_ext = 0 if side in ["left", "both"] else NANOSECONDS_PER_MINUTE
    # NOTE: Add an extra minute to ending boundaries (break_start and close)
    # so we include the last bar (arange doesn't include its stop).
    end_ext = NANOSECONDS_PER_MINUTE if side in ["right", "both"] else 0

    pieces = []
    for open_time, break_start_time, break_end_time, close_time in zip(
        opens_in_ns, break_starts_in_ns, break_ends_in_ns, closes_in_ns
    ):
        if break_start_time != NP_NAT:
            pieces.append(
                np.arange(
                    open_time + start_ext,
                    break_start_time + end_ext,
                    NANOSECONDS_PER_MINUTE,
                )
            )
            pieces.append(
                np.arange(
                    break_end_time + start_ext,
                    close_time + end_ext,
                    NANOSECONDS_PER_MINUTE,
                )
            )
        else:
            pieces.append(
                np.arange(
                    open_time + start_ext,
                    close_time + end_ext,
                    NANOSECONDS_PER_MINUTE,
                )
            )
    out = np.concatenate(pieces).view("datetime64[ns]")
    return out


def one_minute_earlier(arr: np.ndarray) -> np.ndarray:
    """Return an array of nanos one minute behind a given array."""
    arr = arr.copy()
    arr[arr != NP_NAT] -= NANOSECONDS_PER_MINUTE
    return arr


def one_minute_later(arr: np.ndarray) -> np.ndarray:
    """Return an array of nanos one minute ahead of a given array."""
    arr = arr.copy()
    arr[arr != NP_NAT] += NANOSECONDS_PER_MINUTE
    return arr


def is_date(ts: pd.Timestamp) -> bool:
    """Query if a timestamp represents a date (as opposed to a time).

    `ts` considered to represent a date if tz-naive and has time component
    as 00:00.

    Parameters
    ----------
    ts
        Timestamp to query.

    Returns
    -------
    bool
        Boolean indicating if `ts` represents a date.
    """
    return ts.tz is None and ts == ts.normalize()


def to_utc(ts: pd.Timestamp) -> pd.Timestamp:
    """Return a copy of a given timestamp with timezone set to UTC.

    If `ts` is tz-aware will convert `ts` to UTC.
    If `ts` is tz-naive will localize as UTC.

    Parameters
    ----------
    ts
        Timestamp to return a copy of with timezone set to UTC.
    """
    try:
        return ts.tz_convert(pytz.UTC)
    except TypeError:
        return ts.tz_localize(pytz.UTC)


def parse_timestamp(
    timestamp: Date | Minute,
    param_name: str = "minute",
    calendar: ExchangeCalendar | None = None,
    raise_oob: bool = True,
    side: Literal["left", "right", "both", "neither"] | None = None,
    utc: bool = True,
) -> pd.Timestamp:
    """Parse input intended to represent either a date or a minute.

    Parameters
    ----------
    timestamp
        Input to be parsed as either a Date or a Minute. Must be valid
        input to pd.Timestamp.

    param_name
        Name of a parameter that was to receive a Date or Minute.

    calendar
        ExchangeCalendar against which to evaluate out-of-bounds
        timestamps. Only requried if `raise_oob` True of if relying on
        `calendar` for `side`.

    raise_oob : default: True
        True to raise MinuteOutOfBounds if `timestamp` is earlier than the
        first trading minute or later than the last trading minute of
        `calendar`. Pass as False if `timestamp` represents a Date (as
        opposed to a Minute). If True then `calendar` must be passed.

    side : optional, {None, 'left', 'right', 'both', 'neither'}
        The side that determines which minutes at a session's bounds are
        considered as trading minutes (as `ExchangeCalendar` 'side'
        parameter). Only required if `calendar` is not passed or if do not
        wish to rely on `calendar.side`. Ignored if `timestamp` is accurate
        to minute resolution.

    utc : default: True
        True - convert / set timezone to "UTC".
        False - leave any timezone unchanged. Note, if timezone of
        `timestamp` is "UTC" then will remain as "UTC".

    Raises
    ------
    TypeError
        If `timestamp` is not of type [pd.Timestamp | str | int | float |
            datetime.datetime].

    ValueError
        If `timestamp` is not an acceptable single-argument input to
        pd.Timestamp.

    exchange_calendars.errors.MinuteOutOfBounds
        If `raise_oob` True and `timestamp` parses to a valid timestamp
        although timestamp is either before `calendar`'s first trading
        minute or after `calendar`'s last trading minute.
    """
    if isinstance(timestamp, pd.Timestamp):
        ts = timestamp
    else:
        try:
            ts = pd.Timestamp(timestamp)
        except Exception as e:
            msg = (
                f"Parameter `{param_name}` receieved as '{timestamp}' although"
                f" a Date or Minute must be passed as a pd.Timestamp or a"
                f" valid single-argument input to pd.Timestamp."
            )
            if isinstance(e, TypeError):
                raise TypeError(msg) from e
            else:
                raise ValueError(msg) from e

    if utc and ts.tz is not pytz.UTC:
        ts = ts.tz_localize("UTC") if ts.tz is None else ts.tz_convert("UTC")

    if ts.second or ts.microsecond or ts.nanosecond:
        if side is None and calendar is None:
            raise ValueError(
                "`side` or `calendar` must be passed if `timestamp` has a"
                " non-zero second (or more accurate) component. `timestamp`"
                f" parsed as '{ts}'."
            )
        side = side if side is not None else calendar.side
        if side == "left":
            ts = ts.floor("T")
        elif side == "right":
            ts = ts.ceil("T")
        else:
            raise ValueError(
                "`timestamp` cannot have a non-zero second (or more accurate)"
                f" component for `side` '{side}'. `timestamp` parsed as '{ts}'."
            )

    if raise_oob:
        if calendar is None:
            raise ValueError("`calendar` must be passed if `raise_oob` is True.")
        if calendar._minute_oob(ts):
            raise errors.MinuteOutOfBounds(calendar, ts, param_name)

    return ts


def parse_date_or_minute(
    ts: Date | Minute,
    param_name: str,
    calendar: ExchangeCalendar,
) -> tuple[pd.Timestamp, bool]:
    """Parse input that can be interpreted as a Date or a Minute.

    Parameters
    ----------
    ts
        Input that can be interpreted as a Date or a Minute. Must be valid
        input to pd.Timestamp.

    param_name
        Name of a parameter that was to receive a Date or a Minute.

    calendar
        ExchangeCalendar against which to evaluate out-of-bounds
        timestamps.

    Returns
    -------
    tuple[pd.Timestamp, bool]:
        [0] Parsed input to a pd.Timestamp.
        [1] Boolean indicating if input was interpreted as a Minute (True)
            or a Date (False).
    """
    ts = parse_timestamp(ts, param_name, calendar, raise_oob=False, utc=False)
    is_time = not is_date(ts)
    ts = to_utc(ts) if is_time else ts
    if is_time and calendar._minute_oob(ts):
        raise errors.MinuteOutOfBounds(calendar, ts, param_name)
    elif not is_time and calendar._date_oob(ts):
        raise errors.DateOutOfBounds(calendar, ts, param_name)
    return ts, is_time


def parse_trading_minute(
    calendar: ExchangeCalendar, minute: TradingMinute, param_name: str = "minute"
) -> pd.Timestamp:
    """Parse input intended to represent a trading minute.

    Parameters
    ----------
    calendar
       Calendar which `minute` must be a trading minute of.

    minute
        Input to be parsed as a trading minute. Must be valid input to
        pd.Timestamp and represent a trading minute of `calendar`.

    param_name
        Name of a parameter that was to receive a trading minute.

    Raises
    ------
    Errors as `parse_timestamp` and additionally:

    exchange_calendars.errors.NotTradingMinuteError
        If `minute` parses to a valid timestamp although timestamp does not
        represent a trading minute of `calendar`.
    """
    # let out-of-bounds be handled by more specific NotTradingMinuteError message.
    minute = parse_timestamp(minute, param_name, raise_oob=False, calendar=calendar)
    if calendar._minute_oob(minute) or not calendar.is_trading_minute(
        minute, _parse=False
    ):
        raise errors.NotTradingMinuteError(calendar, minute, param_name)
    return minute


def parse_date(
    date: Date,
    param_name: str = "date",
    calendar: ExchangeCalendar | None = None,
    raise_oob: bool = True,
) -> pd.Timestamp:
    """Parse input intended to represent a date.

     Parameters
     ----------
     date
         Input to be parsed as date. Must be valid input to pd.Timestamp
         and have a time component of 00:00.

     param_name
         Name of a parameter that was to receive a date.

    calendar
        ExchangeCalendar against which to evalute out-of-bounds dates.
        Only requried if `raise_oob` True.

    raise_oob : default: True
        True to raise DateOutOfBounds if `date` is earlier than the
        first session or later than the last session of `calendar`. NB if
        True (default) then `calendar` must be passed.

    Returns
     -------
     pd.Timestamp
         pd.Timestamp (timezone naive with time component of 00:00).

     Raises
     ------
     Errors as `parse_timestamp` and additionally:

     ValueError
         If `date` time component is not 00:00.
         If `date` is timezone aware.

    exchange_calendars.errors.DateOutOfBounds
        If `raise_oob` True and `date` parses to a valid timestamp although
        timestamp is before `calendar`'s first session or after
        `calendar`'s last session.
    """
    # side "left" to get it through 'second' handling. Has undesirable effect of
    # allowing `date` to be defined with a second (or more accurate) compoment
    # if it falls within the minute that follows midnight.
    ts = parse_timestamp(date, param_name, raise_oob=False, side="left", utc=False)

    if ts.tz is not None:
        raise ValueError(
            f"Parameter `{param_name}` received with timezone defined as '{ts.tz.zone}'"
            f" although a Date must be timezone naive."
        )

    if not ts == ts.normalize():
        raise ValueError(
            f"Parameter `{param_name}` parsed as '{ts}' although a Date must have"
            f" a time component of 00:00."
        )

    if raise_oob:
        if calendar is None:
            raise ValueError("`calendar` must be passed if `raise_oob` is True.")
        if calendar._date_oob(ts):
            raise errors.DateOutOfBounds(calendar, ts, param_name)

    return ts


def parse_session(
    calendar: ExchangeCalendar, session: Session, param_name: str = "session"
) -> pd.Timestamp:
    """Parse input intended to represent a session label.

    Parameters
    ----------
    calendar
        Calendar against which to evaluate `session`.

    session
        Input to be parsed as session. Must be valid input to pd.Timestamp,
        have a time component of 00:00 and represent a session of
        `calendar`.

    param_name
        Name of a parameter that was to receive a session.

    Returns
    -------
    pd.Timestamp
        pd.Timestamp (timezone naive and with time component of 00:00) that
        represents a real session of `calendar`.

    Raises
    ------
    Errors as `parse_date` and additionally:

    exchange_calendars.errors.NotSessionError
        If `session` parses to a valid date although date does not
        represent a session of `calendar`.
    """
    # let out-of-bounds be handled by more specific NotSessionError message.
    ts = parse_date(session, param_name, raise_oob=False)
    if calendar._date_oob(ts) or not calendar.is_session(ts, _parse=False):
        raise errors.NotSessionError(calendar, ts, param_name)
    return ts


class _TradingIndex:
    """Create a trading index.

    Credit to @Stryder-Git at pandas_market_calendars for showing the way
    with a vectorised solution to creating trading indices.

    Parameters
    ----------
    All parameters as ExchangeCalendar.trading_index
    """

    def __init__(
        self,
        calendar: ExchangeCalendar,
        start_: Date | Minute,
        end_: Date | Minute,
        period: pd.Timedelta,
        # TODO Literal["left", "right", "both", "neither"] when min python 3.8...
        closed: str,
        force_close: bool,
        force_break_close: bool,
        curtail_overlaps: bool,
        ignore_breaks: bool,
    ):
        self.closed = closed
        self.force_break_close = False if ignore_breaks else force_break_close
        self.force_close = force_close
        self.curtail_overlaps = curtail_overlaps

        # parse `start_` and `end_`
        start_, self.start_as_time = parse_date_or_minute(start_, "start", calendar)
        end_, self.end_as_time = parse_date_or_minute(end_, "end", calendar)
        self.start_, self.end_ = start_, end_
        # define start and end as sessions
        start = (
            calendar.minute_to_session(start_, "next") if self.start_as_time else start_
        )
        end = calendar.minute_to_session(end_, "previous") if self.end_as_time else end_

        # get session bound values over requested range
        slice_start = calendar.sessions.searchsorted(start)
        slice_end = calendar.sessions.searchsorted(end, side="right")
        slce = slice(slice_start, slice_end)

        self.interval_nanos = period.value
        self.dtype = np.int64 if self.interval_nanos < 3000000000 else np.int32

        self.opens = calendar.opens_nanos[slce]
        self.closes = calendar.closes_nanos[slce]

        if ignore_breaks:
            self.has_break = False
        else:
            self.break_starts = calendar.break_starts_nanos[slce]
            self.break_ends = calendar.break_ends_nanos[slce]

            self.mask = self.break_starts != pd.NaT.value  # break mask
            self.has_break = self.mask.any()

        self.defaults = {
            "closed": self.closed,
            "force_close": self.force_close,
            "force_break_close": self.force_break_close,
        }

    @property
    def closed_right(self) -> bool:
        return self.closed in ["right", "both"]

    @property
    def closed_left(self) -> bool:
        return self.closed in ["left", "both"]

    def verify_non_overlapping(self):
        """Raise IndicesOverlapError if indices will overlap."""
        if not self.closed_right:
            return

        def _check(
            start_nanos: np.ndarray, end_nanos: np.ndarray, next_start_nanos: np.ndarray
        ):
            """Raise IndicesOverlap Error if indices would overlap.

            `next_start_nanos` describe start of (sub)session that follows and could
            overlap with (sub)session described by `start_nanos` and `end_nanos`.

            All inputs should be of same length.
            """
            num_intervals = np.ceil((end_nanos - start_nanos) / self.interval_nanos)
            right = start_nanos + num_intervals * self.interval_nanos
            if self.closed == "right" and (right > next_start_nanos).any():
                raise errors.IndicesOverlapError()
            if self.closed == "both" and (right >= next_start_nanos).any():
                raise errors.IndicesOverlapError()

        if self.has_break:
            if not self.force_break_close:
                _check(
                    self.opens[self.mask],
                    self.break_starts[self.mask],
                    self.break_ends[self.mask],
                )

        if not self.force_close:
            opens, closes, next_opens = (
                self.opens[:-1],
                self.closes[:-1],
                self.opens[1:],
            )
            _check(opens, closes, next_opens)
            if self.has_break:
                mask = self.mask[:-1]
                _check(self.break_ends[:-1][mask], closes[mask], next_opens[mask])

    def _create_index_for_sessions(
        self,
        start_nanos: np.ndarray,
        end_nanos: np.ndarray,
        force_close: bool,
    ) -> np.ndarray:
        """Create nano array of indices for sessions of given bounds."""
        if start_nanos.size == 0:
            return start_nanos

        # evaluate number of indices for each session
        num_intervals = (end_nanos - start_nanos) / self.interval_nanos
        num_indices = np.ceil(num_intervals).astype("int")

        if force_close:
            if self.closed_right:
                on_freq = (num_intervals == num_indices).all()
                if not on_freq:
                    num_indices -= 1  # add the close later
            else:
                on_freq = True

        if self.closed == "both":
            num_indices += 1
        elif self.closed == "neither":
            num_indices -= 1

        # by session, evaluate a range of int such that indices of a session
        # could be evaluted from [ session_open + (freq * i) for i in range ]
        start = 0 if self.closed_left else 1
        func = np.vectorize(lambda stop: np.arange(start, stop), otypes=[np.ndarray])
        stop = num_indices if self.closed_left else num_indices + 1
        ranges = np.concatenate(func(stop), axis=0, dtype=self.dtype)

        # evaluate index as nano array
        base = start_nanos.repeat(num_indices)
        index = base + ranges * self.interval_nanos

        if force_close and not on_freq:
            index = np.concatenate((index, end_nanos))
            index.sort()

        return index

    def _trading_index(self) -> np.ndarray:
        """Create trading index as nano array.

        Notes
        -----
        If `self.has_break` then index is returned UNSORTED. Why?
        Returning unsorted allows `trading_index_intervals` to create
        indices for the left and right sides and then arrange the right
        in the same order as the sorted left. Although as required, there
        are rare circumstances in which the resulting right side will not
        be in ascending order (it will later be curtailed or an error
        raised). This can happen when, for example, a calendar has breaks,
        `force_break_close` is False although `force_close` is True and the
        period is sufficiently long that the right side of the last
        interval of a morning subsession exceeds the day close, i.e.
        exceeds the right side of the subsequent interval. In these cases,
        sorting the right index by value would result in the indices
        becoming unsynced with the corresponding left indices.
        """
        if self.has_break:

            # sessions with breaks
            index_am = self._create_index_for_sessions(
                self.opens[self.mask],
                self.break_starts[self.mask],
                self.force_break_close,
            )

            index_pm = self._create_index_for_sessions(
                self.break_ends[self.mask], self.closes[self.mask], self.force_close
            )

            # sessions without a break
            index_day = self._create_index_for_sessions(
                self.opens[~self.mask], self.closes[~self.mask], self.force_close
            )

            # put it all together
            index = np.concatenate((index_am, index_pm, index_day))

        else:
            index = self._create_index_for_sessions(
                self.opens, self.closes, self.force_close
            )

        return index

    def curtail_for_times(
        self, index: pd.DatetimeIndex | pd.IntervalIndex
    ) -> pd.DatetimeIndex | pd.IntervalIndex:
        """Curtail start and end of trading index.

        Curtails any unwanted rows from the start and end of `index` if
        class received `start_` and/or `end_` as times.
        """
        intervals = isinstance(index, pd.IntervalIndex)
        bv: np.ndarray | None = None
        if self.start_as_time:
            bv = index.left >= self.start_ if intervals else index >= self.start_
        if self.end_as_time:
            bv_end = index.right <= self.end_ if intervals else index <= self.end_
            bv = bv & bv_end if bv is not None else bv_end
        return index if bv is None else index[bv]

    def trading_index(self) -> pd.DatetimeIndex:
        """Create trading index as a DatetimeIndex."""
        self.verify_non_overlapping()
        index = self._trading_index()
        if self.has_break:
            index.sort()
        index = pd.DatetimeIndex(index, tz="UTC")
        return self.curtail_for_times(index)

    @contextlib.contextmanager
    def _override_defaults(self, **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        yield
        for k, v in self.defaults.items():
            setattr(self, k, v)

    def trading_index_intervals(self) -> pd.IntervalIndex:
        """Create trading index as a pd.IntervalIndex."""
        with self._override_defaults(
            closed="left", force_close=False, force_break_close=False
        ):
            left = self._trading_index()

        if not (self.force_close or self.force_break_close):
            if self.has_break:
                left.sort()
            right = left + self.interval_nanos
        else:
            with self._override_defaults(closed="right"):
                right = self._trading_index()
            if self.has_break:
                # See _trading_index.__doc__ for note on what's going on here.
                indices = left.argsort()
                left.sort()
                right = right[indices]

        overlaps_next = right[:-1] > left[1:]
        if overlaps_next.any():
            if self.curtail_overlaps:
                right[:-1][overlaps_next] = left[1:][overlaps_next]
            else:
                raise errors.IntervalsOverlapError()

        left = pd.DatetimeIndex(left, tz="UTC")
        right = pd.DatetimeIndex(right, tz="UTC")
        index = pd.IntervalIndex.from_arrays(left, right, self.closed)
        return self.curtail_for_times(index)



from __future__ import annotations

import datetime

import numpy as np
import pandas as pd
import pytz
from pytz import UTC


def days_at_time(
    dates: pd.DatetimeIndex,
    time: datetime.time | None,
    tz: pytz.tzinfo.BaseTzInfo,
    day_offset: int,
) -> pd.DatetimeIndex:
    """Return UTC DatetimeIndex of given dates at a given time.

    Parameters
    ----------
    dates
        Dates or date (timezone naive with no time component).

    time
        The time to apply as an offset to each day in `dates`.

    tz
        The timezone in which to interpret `time`.

    day_offset
        Number of days by which to offset each date in `dates`.

    Returns
    -------
    pd.DatetimeIndex
        DatetimeIndex comprising Timestamp evaluted from `dates` and `time`
        with `dates` offset by `day_offset` and `time` interpreted as having
        timezone `tz`. DatetimeIndex has UTC timezone.

    Examples
    --------
    In the example below, the times switch from 13:45 to 12:45 UTC because
    March 13th is the daylight savings transition for America/New_York. All
    the times are still 8:45 when interpreted in America/New_York.

    >>> import pandas as pd; import datetime; import pprint
    >>> dts = pd.date_range('2016-03-12', '2016-03-14')
    >>> dts_845 = days_at_time(dts, datetime.time(8, 45), 'America/New_York', 0)
    >>> pprint.pprint([str(dt) for dt in dts_845])
    ['2016-03-12 13:45:00+00:00',
     '2016-03-13 12:45:00+00:00',
     '2016-03-14 12:45:00+00:00']
    """
    if time is None:
        return pd.DatetimeIndex([None for _ in dates]).tz_localize(UTC)

    if len(dates) == 0:
        return dates.tz_localize(UTC)

    # Offset days without tz to avoid timezone issues.
    delta = pd.Timedelta(
        days=day_offset,
        hours=time.hour,
        minutes=time.minute,
        seconds=time.second,
    )
    return (dates + delta).tz_localize(tz).tz_convert(UTC)


def vectorized_sunday_to_monday(dtix):
    """A vectorized implementation of
    :func:`pandas.tseries.holiday.sunday_to_monday`.

    Parameters
    ----------
    dtix : pd.DatetimeIndex
        The index to shift sundays to mondays.

    Returns
    -------
    sundays_as_mondays : pd.DatetimeIndex
        ``dtix`` with all sundays moved to the next monday.
    """
    values = dtix.values.copy()
    values[dtix.weekday == 6] += np.timedelta64(1, "D")
    return pd.DatetimeIndex(values)


def longest_run(ser: pd.Series) -> pd.Index:
    """Get the longest run of consecutive True values in a Series.

    Function can be used to find the longest run of values that meet a
    condition.

    Parameters
    ----------
    ser
        pd.Series of bool dtype.
            Index should reflect values against which a condition was
                assessed.
            Values should reflect whether corresponding index value
                met the condition.

    Return
    ------
    pd.Index
        Slice of `ser` index that corresponds with the longest run of
            consecutive True values.

    Examples
    --------
    >>> arr = np.arange(0, 88)
    >>> ser = pd.Series(arr, index=arr)
    >>> bv = (
    ...     ((ser >= 10) & (ser < 16))
    ...     | ((ser >= 30) & (ser <= 40))
    ...     | ((ser >= 55) & (ser < 61))
    ... )
    >>> longest_run(bv)
    Int64Index([30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40], dtype='int64')
    >>> pd.testing.assert_index_equal(longest_run(bv), ser.index[30:41])
    """
    # group Trues by only adding to sum when value False.
    trues_grouped = (~ser).cumsum()[ser]  # and only take True Values
    group_sizes = trues_grouped.value_counts()  # count each run
    max_run_size = group_sizes.max()
    max_run_group_id = group_sizes[group_sizes == max_run_size].index[0]
    run = trues_grouped[trues_grouped == max_run_group_id].index
    return run


def indexes_union(indexes: list[pd.Index]) -> pd.Index:
    """Return union of multiple pd.Index objects.

    Parameters
    ----------
    indexes
        Index objects to be joined. All indexes must be of same dtype.

    Examples
    --------
    >>> index1 = pd.date_range('2021-05-01 12:20', periods=2, freq='1H')
    >>> index2 = pd.date_range('2021-05-02 17:10', periods=2, freq='22T')
    >>> index3 = pd.date_range('2021-05-03', periods=2, freq='1D')
    >>> indexes_union([index1, index2, index3])
    DatetimeIndex(['2021-05-01 12:20:00', '2021-05-01 13:20:00',
                   '2021-05-02 17:10:00', '2021-05-02 17:32:00',
                   '2021-05-03 00:00:00', '2021-05-04 00:00:00'],
                  dtype='datetime64[ns]', freq=None)
    """
    index = indexes[0]
    for indx in indexes[1:]:
        index = index.union(indx)
    return index



from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import (
    MO,
    DateOffset,
    GoodFriday,
    Holiday,
    weekend_to_monday,
)
from pytz import timezone

from .common_holidays import (
    boxing_day,
    christmas,
    new_years_day,
    weekend_boxing_day,
    weekend_christmas,
)
from .exchange_calendar import (
    FRIDAY,
    MONDAY,
    THURSDAY,
    TUESDAY,
    WEDNESDAY,
    HolidayCalendar,
    ExchangeCalendar,
)

# New Year's Day
XTSENewYearsDay = new_years_day(observance=weekend_to_monday)

# Ontario Family Day
FamilyDay = Holiday(
    "Family Day",
    month=2,
    day=1,
    offset=DateOffset(weekday=MO(3)),
    start_date="2008-01-01",
)
# Victoria Day
VictoriaDay = Holiday(
    "Victoria Day",
    month=5,
    day=24,
    offset=DateOffset(weekday=MO(-1)),
)
# Canada Day
CanadaDay = Holiday(
    "Canada Day",
    month=7,
    day=1,
    observance=weekend_to_monday,
)
# Civic Holiday
CivicHoliday = Holiday(
    "Civic Holiday",
    month=8,
    day=1,
    offset=DateOffset(weekday=MO(1)),
)
# Labor Day
LaborDay = Holiday(
    "Labor Day",
    month=9,
    day=1,
    offset=DateOffset(weekday=MO(1)),
)
# Canadian Thanksgiving
CanadianThanksgiving = Holiday(
    "Canadian Thanksgiving",
    month=10,
    day=1,
    offset=DateOffset(weekday=MO(2)),
)

ChristmasEveEarlyClose2010Onwards = Holiday(
    "Christmas Eve Early Close",
    month=12,
    day=24,
    days_of_week=(MONDAY, TUESDAY, WEDNESDAY, THURSDAY, FRIDAY),
    start_date=pd.Timestamp("2010-01-01"),
)

Christmas = christmas()

WeekendChristmas = weekend_christmas()

BoxingDay = boxing_day()

WeekendBoxingDay = weekend_boxing_day()

September11ClosingsCanada = pd.date_range("2001-09-11", "2001-09-12")


class XTSEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Toronto Stock Exchange (XTSE).

    Open Time: 9:30 AM, EST
    Close Time: 4:00 PM, EST

    Regularly-Observed Holidays:
    - New Years Day (observed on first business day on/after)
    - Family Day (Third Monday in February, starting in 2008)
    - Good Friday
    - Victoria Day (Monday before May 25th)
    - Canada Day (July 1st, observed first business day after)
    - Civic Holiday (First Monday in August)
    - Labor Day (First Monday in September)
    - Thanksgiving (Second Monday in October)
    - Christmas Day
        - Dec. 26th if Christmas is on a Sunday
        - Dec. 27th if Christmas is on a weekend
    - Boxing Day
        - Dec. 27th if Christmas is on a Sunday
        - Dec. 28th if Boxing Day is on a weekend

    Early closes:
    - Starting in 2010, if Christmas Eve falls on a weekday, the market
      closes at 1:00 pm that day. If it falls on a weekend, there is no
      early close.
    """

    regular_early_close = time(13)

    name = "XTSE"

    tz = timezone("America/Toronto")

    open_times = ((None, time(9, 30)),)

    close_times = ((None, time(16)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                XTSENewYearsDay,
                FamilyDay,
                GoodFriday,
                VictoriaDay,
                CanadaDay,
                CivicHoliday,
                LaborDay,
                CanadianThanksgiving,
                Christmas,
                WeekendChristmas,
                BoxingDay,
                WeekendBoxingDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        # NOTE: change the name of this property
        return list(chain(September11ClosingsCanada))

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar([ChristmasEveEarlyClose2010Onwards]),
            )
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import GoodFriday, Holiday
from pytz import timezone

from .common_holidays import (
    all_saints_day,
    christmas,
    european_labour_day,
    immaculate_conception,
    maundy_thursday,
    new_years_day,
    new_years_eve,
    saint_peter_and_saint_paul_day,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

####################
# Regular Holidays #
####################
NewYearsDay = new_years_day()

MaundyThursday = maundy_thursday()

LabourDay = european_labour_day()

SaintPeterAndSaintPaulDay = saint_peter_and_saint_paul_day()

IndependenceDay1 = Holiday("Independence Day", month=7, day=28)
IndependenceDay2 = Holiday("Independence Day", month=7, day=29)

SantaRosa = Holiday("Santa Rosa", month=8, day=30)

BattleOfAngamos = Holiday("Battle of Angamos", month=10, day=8)

AllSaintsDay = all_saints_day()

ImmaculateConception = immaculate_conception()

Christmas = christmas()

NewYearsEve = new_years_eve(end_date="2008")


##################
# Adhoc Holidays #
##################
ExchangeHolidays = [
    pd.Timestamp("2009-01-02"),
    pd.Timestamp("2009-07-27"),
    pd.Timestamp("2015-07-27"),
    pd.Timestamp("2015-10-09"),
]

NationalHolidays = [
    pd.Timestamp("2015-01-02"),
]

ASPASummit = [
    pd.Timestamp("2012-10-01"),
    pd.Timestamp("2012-10-02"),
]

APECSummit = [
    pd.Timestamp("2016-11-17"),
    pd.Timestamp("2016-11-18"),
]

EighthSummitOfTheAmericas = [pd.Timestamp("2018-04-13")]


class XLIMExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Lima Stock Exchange (Bolsa de Valores de Lima) in Peru.

    Open Time: 9:00 AM, Peruvian Time
    Close Time: 4:00 PM, Peruvian Time

    Regularly-Observed Holidays:
      - New Year's Day
      - Maundy Thursday
      - Good Friday
      - Labour Day
      - Saint Paul and Saint Peter Day
      - Independence Day
      - Santa Rosa
      - Battle of Angamos
      - All Saints' Day
      - Immaculate Conception
      - Christmas Day

    Holidays No Longer Observed:
      - New Year's Eve

    Early Closes:
      - None
    """

    name = "XLIM"
    tz = timezone("America/Lima")

    open_times = ((None, time(9)),)
    close_times = ((None, time(16)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                MaundyThursday,
                GoodFriday,
                LabourDay,
                SaintPeterAndSaintPaulDay,
                IndependenceDay1,
                IndependenceDay2,
                SantaRosa,
                BattleOfAngamos,
                AllSaintsDay,
                ImmaculateConception,
                Christmas,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return list(
            chain(
                ExchangeHolidays,
                NationalHolidays,
                ASPASummit,
                APECSummit,
                EighthSummitOfTheAmericas,
            )
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from datetime import time
import functools

import pandas as pd
from pandas.tseries.holiday import Holiday
from pandas.tseries.offsets import CustomBusinessDay
import pytz

from .exchange_calendar import HolidayCalendar
from .precomputed_exchange_calendar import PrecomputedExchangeCalendar
from .xkrx_holidays import (
    krx_regular_holiday_rules,
    precomputed_krx_holidays,
    precomputed_csat_days,
)
from .pandas_extensions.offsets import MultipleWeekmaskCustomBusinessDay
from .pandas_extensions.korean_holiday import next_business_day


class XKRXExchangeCalendar(PrecomputedExchangeCalendar):
    """
    Calendar for the Korea exchange, and the primary calendar for
    the country of South Korea.

    Open Time: 9:00 AM, KST (Korean Standard Time)
    Close Time: 3:30 PM, KST (Korean Standard Time)

    NOTE: Korea observes Standard Time year-round.

    Due to the complexity around the Korean holidays, we are hardcoding
    a list of holidays covering 1986-2019, inclusive.

    Regularly-Observed Holidays:
    - Seollal (New Year's Day)
    - Independence Movement Day
    - Labor Day
    - Buddha's Birthday
    - Memorial Day
    - Provincial Election Day
    - Liberation Day
    - Chuseok (Korean Thanksgiving)
    - National Foundation Day
    - Christmas Day
    - End of Year Holiday

    NOTE: Hangeul Day became a national holiday in 2013
    - Hangeul Proclamation Day
    """

    name = "XKRX"

    tz = pytz.timezone("Asia/Seoul")

    # KRX schedule change history
    # https://blog.naver.com/daishin_blog/220724111002

    # 1956-03-03: 0930~1130, 1330~1530
    # 1978-04-??: 1000~1200, 1330~1530
    # 1986-04-??: 0940~1200, 1320~1520
    # 1987-03-??: 0940~1140, 1320~1520
    # 1995-01-01: 0930~1130, 1300~1500
    # 1998-12-07: 0900~1200, 1300~1500
    # 2000-05-22: 0900~1500
    # 2016-08-01: 0900~1530

    # Break time disappears since 2000-05-22
    # https://www.donga.com/news/Economy/article/all/20000512/7534650/1

    # Closing time became 30mins late since 2016-08-01
    # https://biz.chosun.com/site/data/html_dir/2016/07/24/2016072400309.html

    open_times = (
        (None, time(9, 30)),
        (pd.Timestamp("1978-04-01"), time(10, 0)),
        (pd.Timestamp("1986-04-01"), time(9, 40)),
        (pd.Timestamp("1995-01-01"), time(9, 30)),
        (pd.Timestamp("1998-12-07"), time(9, 0)),
    )
    break_start_times = (
        (None, time(11, 30)),
        (pd.Timestamp("1978-04-01"), time(12, 0)),
        (pd.Timestamp("1987-03-01"), time(11, 40)),
        (pd.Timestamp("1995-01-01"), time(11, 30)),
        (pd.Timestamp("1998-12-07"), time(12, 0)),
        (pd.Timestamp("2000-05-22"), None),
    )
    break_end_times = (
        (None, time(13, 30)),
        (pd.Timestamp("1986-04-01"), time(13, 20)),
        (pd.Timestamp("1995-01-01"), time(13, 0)),
        (pd.Timestamp("2000-05-22"), None),
    )
    close_times = (
        (None, time(15, 30)),
        (pd.Timestamp("1986-04-01"), time(15, 20)),
        (pd.Timestamp("1995-01-01"), time(15, 0)),
        (pd.Timestamp("2016-08-01"), time(15, 30)),
    )

    # Saterday became holiday since 1998-12-07
    # https://www.hankyung.com/finance/article/1998080301961

    weekmask = "1111100"

    @property
    def special_weekmasks(self):
        """
        Returns
        -------
        list: List of (date, date, str) tuples that represent special
         weekmasks that applies between dates.
        """
        return [
            (None, pd.Timestamp("1998-12-06"), "1111110"),
        ]

    @classmethod
    def precomputed_holidays(cls) -> list[pd.Timestamp]:
        return precomputed_krx_holidays.tolist()

    @classmethod
    def _earliest_precomputed_year(cls) -> int:
        return 1956

    @classmethod
    def _latest_precomputed_year(cls) -> int:
        return 2050

    # KRX regular and precomputed adhoc holidays

    @property
    def regular_holidays(self):
        return HolidayCalendar(krx_regular_holiday_rules)

    # The first business day of each year:
    #  opening schedule is delayed by an hour.

    @property
    def special_offsets(self):
        """
        Returns
        -------
        list: List of (timedelta, timedelta, timedelta, timedelta, AbstractHolidayCalendar) tuples
         that represent special open, break_start, break_end, close offsets
         and corresponding HolidayCalendars.
        """
        return [
            (
                pd.Timedelta(1, unit="h"),
                None,
                None,
                None,
                HolidayCalendar(
                    [
                        Holiday(
                            "First Business Day of Year",
                            month=1,
                            day=1,
                            observance=next_business_day,
                        )
                    ]
                ),
            ),
        ]

    # Every year's CSAT day, all schedules are delayed by:
    #  before 1998-11-18: 30 minutes
    #  after  1998-11-18: 1 hour

    @property
    def special_offsets_adhoc(
        self,
    ) -> list[
        tuple[pd.Timedelta, pd.Timedelta, pd.Timedelta, pd.Timedelta, pd.DatetimeIndex]
    ]:
        """
        Returns
        -------
        list: List of (timedelta, timedelta, timedelta, timedelta, DatetimeIndex) tuples
         that represent special open, break_start, break_end, close offsets
         and corresponding DatetimeIndexes.
        """
        return [
            (
                pd.Timedelta(30, unit="m"),
                pd.Timedelta(30, unit="m"),
                pd.Timedelta(30, unit="m"),
                pd.Timedelta(30, unit="m"),
                precomputed_csat_days[
                    precomputed_csat_days.slice_indexer("1993-08-20", "1998-11-17")
                ],
            ),
            (
                pd.Timedelta(1, unit="h"),
                pd.Timedelta(1, unit="h"),
                pd.Timedelta(1, unit="h"),
                pd.Timedelta(1, unit="h"),
                precomputed_csat_days[
                    precomputed_csat_days.slice_indexer("1998-11-18", None)
                ],
            ),
        ]

    def _overwrite_special_offsets(
        self,
        session_labels: pd.DatetimeIndex,
        standard_times: pd.DatetimeIndex | None,
        offsets: tuple[pd.Timedelta, HolidayCalendar],
        ad_hoc_offsets: tuple[pd.Timedelta, pd.DatetimeIndex],
        start_date: pd.Timestamp,
        end_date: pd.Timestamp,
        strict: bool = False,
    ):
        # Short circuit when nothing to apply.
        if standard_times is None or not len(standard_times):
            return

        len_m, len_oc = len(session_labels), len(standard_times)
        if len_m != len_oc:
            raise ValueError(
                "Found misaligned dates while building calendar.\nExpected"
                " session_labels to be the same length as open_or_closes but,\n"
                f"len(session_labels)={len_m}, len(open_or_closes)={len_oc}"
            )

        regular = []
        for offset, calendar in offsets:
            days = calendar.holidays(start_date, end_date)
            series = pd.Series(index=days, data=offset)
            regular.append(series)

        ad_hoc = []
        for offset, datetimes in ad_hoc_offsets:
            series = pd.Series(index=datetimes, data=offset)
            ad_hoc.append(series)

        merged = regular + ad_hoc
        if not merged:
            return pd.Series([], dtype="timedelta64[ns]")

        result = pd.concat(merged).sort_index()
        offsets = result.loc[(result.index >= start_date) & (result.index <= end_date)]

        # Find the array indices corresponding to each special date.
        indexer = session_labels.get_indexer(offsets.index)

        # -1 indicates that no corresponding entry was found.  If any -1s are
        # present, then we have special dates that doesn't correspond to any
        # trading day.
        if -1 in indexer and strict:
            bad_dates = list(offsets.index[indexer == -1])
            raise ValueError(f"Special dates {bad_dates} are not trading days.")

        special_opens_or_closes = standard_times[indexer] + offsets

        # Short circuit when nothing to apply.
        if not len(special_opens_or_closes):
            return

        # NOTE: This is a slightly dirty hack.  We're in-place overwriting the
        # internal data of an Index, which is conceptually immutable.  Since we're
        # maintaining sorting, this should be ok, but this is a good place to
        # sanity check if things start going haywire with calendar computations.
        standard_times.values[indexer] = special_opens_or_closes.values

    def apply_special_offsets(
        self,
        session_labels: pd.DatetimeIndex,
        start: pd.Timestamp,
        end: pd.Timestamp,
    ):
        """Evaluate and overwrite special offsets."""
        _special_offsets = self.special_offsets
        _special_offsets_adhoc = self.special_offsets_adhoc

        _special_open_offsets = [
            (t[0], t[-1]) for t in _special_offsets if t[0] is not None
        ]
        _special_open_offsets_adhoc = [
            (t[0], t[-1]) for t in _special_offsets_adhoc if t[0] is not None
        ]
        _special_break_start_offsets = [
            (t[1], t[-1]) for t in _special_offsets if t[1] is not None
        ]
        _special_break_start_offsets_adhoc = [
            (t[1], t[-1]) for t in _special_offsets_adhoc if t[1] is not None
        ]
        _special_break_end_offsets = [
            (t[2], t[-1]) for t in _special_offsets if t[2] is not None
        ]
        _special_break_end_offsets_adhoc = [
            (t[2], t[-1]) for t in _special_offsets_adhoc if t[2] is not None
        ]
        _special_close_offsets = [
            (t[3], t[-1]) for t in _special_offsets if t[3] is not None
        ]
        _special_close_offsets_adhoc = [
            (t[3], t[-1]) for t in _special_offsets_adhoc if t[3] is not None
        ]

        self._overwrite_special_offsets(
            session_labels,
            self._opens,
            _special_open_offsets,
            _special_open_offsets_adhoc,
            start,
            end,
        )
        self._overwrite_special_offsets(
            session_labels,
            self._break_starts,
            _special_break_start_offsets,
            _special_break_start_offsets_adhoc,
            start,
            end,
        )
        self._overwrite_special_offsets(
            session_labels,
            self._break_ends,
            _special_break_end_offsets,
            _special_break_end_offsets_adhoc,
            start,
            end,
        )
        self._overwrite_special_offsets(
            session_labels,
            self._closes,
            _special_close_offsets,
            _special_close_offsets_adhoc,
            start,
            end,
        )

    @functools.cached_property
    def day(self):
        if self.special_weekmasks:
            return MultipleWeekmaskCustomBusinessDay(
                holidays=self.adhoc_holidays,
                calendar=self.regular_holidays,
                weekmask=self.weekmask,
                weekmasks=self.special_weekmasks,
            )
        else:
            return CustomBusinessDay(
                holidays=self.adhoc_holidays,
                calendar=self.regular_holidays,
                weekmask=self.weekmask,
            )


class PrecomputedXKRXExchangeCalendar(PrecomputedExchangeCalendar):
    """
    Calendar for the Korea exchange, and the primary calendar for
    the country of South Korea.

    Open Time: 9:00 AM, KST (Korean Standard Time)
    Close Time: 3:30 PM, KST (Korean Standard Time)

    NOTE: Korea observes Standard Time year-round.

    Due to the complexity around the Korean holidays, we are hardcoding
    a list of holidays covering 1986-2019, inclusive.

    Regularly-Observed Holidays:
    - Seollal (New Year's Day)
    - Independence Movement Day
    - Labor Day
    - Buddha's Birthday
    - Memorial Day
    - Provincial Election Day
    - Liberation Day
    - Chuseok (Korean Thanksgiving)
    - National Foundation Day
    - Christmas Day
    - End of Year Holiday

    NOTE: Hangeul Day became a national holiday in 2013
    - Hangeul Proclamation Day
    """

    name = "XKRX"

    tz = pytz.timezone("Asia/Seoul")

    open_times = ((None, time(9)),)
    close_times = ((None, time(15, 30)),)

    @classmethod
    def precomputed_holidays(cls):
        return precomputed_krx_holidays



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas.tseries.holiday import EasterMonday, GoodFriday, Holiday, weekend_to_monday
from pytz import timezone

from .common_holidays import (
    all_saints_day,
    assumption_day,
    boxing_day,
    christmas,
    christmas_eve,
    epiphany,
    european_labour_day,
    immaculate_conception,
    new_years_day,
    new_years_eve,
)
from .exchange_calendar import WEEKDAYS, HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day()

Epiphany = epiphany(end_date="2007")

LabourDay = european_labour_day()

AssumptionDay = assumption_day(end_date="2005", observance=weekend_to_monday)

NationalDay = Holiday(
    "National Day",
    month=10,
    day=12,
    end_date="2005",
)

AllSaintsDay = all_saints_day(end_date="2005")

ConstitutionDay = Holiday(
    "Constitution Day",
    month=12,
    day=6,
    end_date="2005",
)

ImmaculateConception = immaculate_conception(end_date="2005")

ChristmasEveThrough2010 = christmas_eve(end_date="2011")
ChristmasEveFrom2021 = christmas_eve(start_date="2021")
ChristmasEveEarlyClose2012To2020 = christmas_eve(
    start_date="2012",
    end_date="2021",
    days_of_week=(WEEKDAYS),
)
Christmas = christmas()
BoxingDay = boxing_day()

NewYearsEveThrough2010 = new_years_eve(end_date="2011")
NewYearsEveFrom2021 = new_years_eve(start_date="2021")
NewYearsEveEarlyClose2012To2020 = new_years_eve(
    start_date="2012",
    end_date="2021",
    days_of_week=(WEEKDAYS),
)


class XMADExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Madrid Stock Exchange (Bolsa de Madrid).

    Open Time: 9:00 AM, CET (Central European Time)
    Close Time: 5:30 PM, CET (Central European Time)

    Regularly-Observed Holidays:
      - New Year's Day
      - Good Friday
      - Easter Monday
      - Labour Day
      - Christmas Eve (apart from when it was observed as an early close)
      - Christmas Day
      - Boxing Day

    Holidays No Longer Observed:
      - Epiphany (until 2006, inclusive)
      - Assumption Day (until 2004, inclusive)
      - National Day (until 2004, inclusive)
      - All Saints Day (until 2004, inclusive)
      - Constitution Day (until 2004, inclusive)
      - Immaculate Conception (until 2004, inclusive)
      - New Year's Eve (until 2010, inclusive)

    Early Closes:
      - Christmas Eve (2012 to 2020, inclusive)
      - New Year's Eve (2012 and after)
    """

    regular_early_close = time(14, 00)

    name = "XMAD"

    tz = timezone("Europe/Madrid")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                Epiphany,
                GoodFriday,
                EasterMonday,
                LabourDay,
                AssumptionDay,
                NationalDay,
                AllSaintsDay,
                ConstitutionDay,
                ImmaculateConception,
                ChristmasEveThrough2010,
                ChristmasEveFrom2021,
                Christmas,
                BoxingDay,
                NewYearsEveThrough2010,
                NewYearsEveFrom2021
            ]
        )

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        ChristmasEveEarlyClose2012To2020,
                        NewYearsEveEarlyClose2012To2020,
                    ]
                ),
            )
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time
from itertools import chain

import pandas as pd
from pandas.tseries.holiday import MO, DateOffset, GoodFriday, Holiday
from pytz import timezone

from .common_holidays import (
    all_saints_day,
    christmas,
    christmas_eve,
    european_labour_day,
    maundy_thursday,
    new_years_day,
    new_years_eve,
)
from .lunisolar_holidays import chinese_lunar_new_year_dates
from .exchange_calendar import FRIDAY, HolidayCalendar, ExchangeCalendar


def only_friday(dt):
    """
    Only keeps the holidays that fall on a Friday.  Useful
    for defining holidays that add a Friday to make a 4-day weekend
    when falling on a Thursday.
    """
    return dt[dt.weekday == FRIDAY]


# All pre-2011 holidays are pre-computed, so we define Holidays starting
# in 2011.
NewYearsDay = new_years_day(start_date="2011")

# Chinese New Year's only observed after 2011
ChineseNewYear = chinese_lunar_new_year_dates

ChineseNewYearAfter2011 = ChineseNewYear[ChineseNewYear.year > 2011]

PeoplePowerRevolution = Holiday(
    "People Power Revolution",
    month=2,
    day=25,
    start_date="2016",
)

ArawNgKagitingan = Holiday(
    "Araw Ng Kagitingan",
    month=4,
    day=9,
    start_date="2011",
)

MaundyThursday = maundy_thursday(start_date="2011")

LabourDay = european_labour_day(start_date="2011")

IndependenceDay = Holiday(
    "Independence Day",
    month=6,
    day=12,
    start_date="2011",
)

NinoyAquinoDay = Holiday(
    "Ninoy Aquino Day",
    month=8,
    day=21,
    start_date="2011",
)

NationalHeroesDay = Holiday(
    "National Heroes Day",
    month=8,
    day=31,
    offset=DateOffset(weekday=MO(-1)),
    start_date="2011",
)

AllSaintsDay = all_saints_day(start_date="2011")

AllSaintsDayExtra = Holiday(
    "Friday after All Saint's Day",
    month=11,
    day=2,
    start_date="2011",
    observance=only_friday,
)

BonifacioDay = Holiday(
    "Bonifacio Day",
    month=11,
    day=30,
    start_date="2011",
)

ChristmasEve = christmas_eve(start_date="2011")

Christmas = christmas(start_date="2011")

RizalDay = Holiday(
    "Rizal Day",
    month=12,
    day=30,
    start_date="2011",
)

NewYearsEve = new_years_eve(start_date="2011")

# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Philippines' observance of Eid al-Fitr.  Other countries that
# observe Eid al-Fitr might use slightly different dates
philippines_eid_al_fitr = pd.to_datetime(
    [
        "2011-08-30",
        "2012-08-20",
        "2013-08-09",
        "2014-07-29",
        "2015-07-17",
        "2016-07-06",
        "2017-06-26",
        "2018-06-15",
        "2019-06-05",
    ]
)

# These dates were initially calculated using the ummalqura Python
# package (https://pypi.org/project/ummalqura/), and then tweaked
# to fit Philippines' observance of Eid al-Adha.  Other countries that
# observe Eid al-Adha might use slightly different dates
philippines_eid_al_adha = pd.to_datetime(
    [
        "2011-11-07",
        "2012-10-26",
        "2013-10-15",
        "2014-10-06",
        "2015-09-25",
        "2016-09-12",
        "2017-09-01",
        "2018-08-21",
        "2019-08-12",
    ]
)


class XPHSExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Philippine Stock Exchange (XPHS).

    Open Time: 9:30 AM, PHT
    Close Time: 3:30 PM, PHT

    Regularly-Observed Holidays:
    - New Year's Day
    - Chinese New Year's Day
    - Araw Ng Kagitingan (Apr 9)
    - Maundy Thursday (Thurs before Easter)
    - Good Friday (Friday before Easter)
    - Labour Day (May 1)
    - Eid-ul-Fitr (Islamic Lunar Calendar)
    - Independence Day (Jun 12)
    - Eid-al-Adha
    - Ninoy Aquino Day (Aug 21)
    - National Heroes' Day (last Monday of August)
    - All Saint's Day (Nov 1)
    - Bonifacio Day (Nov 30)
    - Christmas Eve (Dec 24)
    - Christmas Day (Dec 25)
    - Rizal Day (Dec 30)
    - New Year's Eve (Dec 31)


    Early Closes:
    - None
    """

    name = "XPHS"

    tz = timezone("Asia/Manila")

    open_times = ((None, time(9, 30)),)

    close_times = ((None, time(15, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                PeoplePowerRevolution,
                ArawNgKagitingan,
                MaundyThursday,
                GoodFriday,
                LabourDay,
                IndependenceDay,
                NinoyAquinoDay,
                NationalHeroesDay,
                AllSaintsDay,
                AllSaintsDayExtra,
                BonifacioDay,
                ChristmasEve,
                Christmas,
                RizalDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        misc_adhoc = [
            "2017-04-28",
            "2017-10-16",
            "2015-11-18",
            "2015-11-19",
            "2013-08-19",
            "2013-08-20",
            "2013-10-28",
            "2011-06-20",
            # Election Day
            "2019-05-13",
            "2018-05-14",
            "2016-05-09",
            "2013-05-13",
            # Typhoon
            "2017-09-12",
            "2014-12-08",
            "2014-09-19",
            "2014-07-16",
            "2012-08-07",
            "2011-09-27",
            # Pope's Visit
            "2015-01-15",
            "2015-01-16",
            # Christmas Makeups
            "2017-12-26",
            "2016-12-26",
            "2014-12-26",
            # New Years Makeups
            "2018-01-02",
            "2017-01-02",
            "2015-01-02",
            # Other Makeups
            "2017-10-31",  # All Saint's Makeup
            "2016-10-31",  # All Saint's Makeup
            "2011-10-31",  # All Saint's Makeup
        ]

        # Before 2011, there are a few holidays that follow makeup day
        # rules that are inconsistent/challenging to model. Starting
        # in 2011, the rules become much more consistent. For convenience
        # we are hardcoding pre-2011 holidays and defining holidays in
        # 2011-present
        pre_2011_holidays = [
            "2002-01-01",
            "2002-02-25",
            "2002-03-28",
            "2002-03-29",
            "2002-04-08",
            "2002-05-01",
            "2002-06-12",
            "2002-07-15",
            "2002-10-31",
            "2002-11-01",
            "2002-12-06",
            "2002-12-24",
            "2002-12-25",
            "2002-12-30",
            "2002-12-31",
            "2003-01-01",
            "2003-02-25",
            "2003-04-07",
            "2003-04-17",
            "2003-04-18",
            "2003-05-01",
            "2003-05-02",
            "2003-06-13",
            "2003-08-22",
            "2003-11-26",
            "2003-12-24",
            "2003-12-25",
            "2003-12-26",
            "2003-12-31",
            "2004-01-01",
            "2004-01-02",
            "2004-02-25",
            "2004-04-07",
            "2004-04-08",
            "2004-04-09",
            "2004-05-10",
            "2004-11-01",
            "2004-11-15",
            "2004-11-29",
            "2004-12-03",
            "2004-12-24",
            "2004-12-27",
            "2004-12-30",
            "2004-12-31",
            "2005-02-25",
            "2005-03-23",
            "2005-03-24",
            "2005-03-25",
            "2005-05-02",
            "2005-06-13",
            "2005-07-25",
            "2005-08-29",
            "2005-10-31",
            "2005-11-01",
            "2005-11-04",
            "2005-11-28",
            "2005-12-26",
            "2005-12-30",
            "2006-04-13",
            "2006-04-14",
            "2006-05-01",
            "2006-06-12",
            "2006-07-24",
            "2006-08-21",
            "2006-09-28",
            "2006-09-29",
            "2006-10-24",
            "2006-11-01",
            "2006-12-01",
            "2006-12-25",
            "2006-12-26",
            "2007-01-01",
            "2007-04-05",
            "2007-04-06",
            "2007-04-09",
            "2007-05-01",
            "2007-05-14",
            "2007-06-11",
            "2007-08-20",
            "2007-08-27",
            "2007-10-12",
            "2007-10-29",
            "2007-11-01",
            "2007-11-02",
            "2007-11-30",
            "2007-12-24",
            "2007-12-25",
            "2007-12-31",
            "2008-01-01",
            "2008-02-25",
            "2008-03-20",
            "2008-03-21",
            "2008-04-07",
            "2008-05-01",
            "2008-06-09",
            "2008-08-18",
            "2008-08-25",
            "2008-10-01",
            "2008-12-01",
            "2008-12-25",
            "2008-12-26",
            "2008-12-29",
            "2008-12-30",
            "2008-12-31",
            "2009-01-01",
            "2009-01-02",
            "2009-04-06",
            "2009-04-09",
            "2009-04-10",
            "2009-05-01",
            "2009-06-12",
            "2009-07-17",
            "2009-08-05",
            "2009-08-21",
            "2009-08-31",
            "2009-09-07",
            "2009-09-21",
            "2009-11-02",
            "2009-11-30",
            "2009-12-24",
            "2009-12-25",
            "2009-12-30",
            "2009-12-31",
            "2010-01-01",
            "2010-04-01",
            "2010-04-02",
            "2010-04-09",
            "2010-05-03",
            "2010-05-10",
            "2010-06-14",
            "2010-06-30",
            "2010-08-30",
            "2010-09-10",
            "2010-10-25",
            "2010-11-01",
            "2010-11-16",
            "2010-11-29",
            "2010-12-24",
            "2010-12-27",
            "2010-12-31",
        ]

        return list(
            chain(
                misc_adhoc,
                pre_2011_holidays,
                ChineseNewYearAfter2011,
                philippines_eid_al_adha,
                philippines_eid_al_fitr,
            )
        )



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from dateutil.relativedelta import MO
from pandas import DateOffset, Timestamp
from pandas.tseries.holiday import (
    EasterMonday,
    GoodFriday,
    Holiday,
    previous_friday,
    sunday_to_monday,
    weekend_to_monday,
)
from pytz import timezone

from .common_holidays import (
    anzac_day,
    boxing_day,
    christmas,
    new_years_day,
    weekend_boxing_day,
    weekend_christmas,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

NewYearsDay = new_years_day(observance=weekend_to_monday)

# "Celebrating Australia: A history of Australia Day" Elizabeth Kwan
# https://www.australiaday.org.au/storage/celebratingaustralia.pdf
# present rule follows the 1994 agreement between the States to
# synchronize the holiday with weekends mondayized
AustraliaDay = Holiday(
    "Australia Day",
    month=1,
    day=26,
    start_date=Timestamp("1994-01-01"),
    observance=weekend_to_monday,
)

# prior to 1993 the holiday was observed on the Monday
# following, or on, the 26th of January
AustraliaDayPre88 = Holiday(
    "Australia Day",
    month=1,
    day=26,
    start_date=Timestamp("1960-01-01"),
    end_date=Timestamp("1987-12-31"),
    offset=DateOffset(weekday=MO(1)),
)
# The 1988 Bi-Centennial celebrations saw an extra holiday
# and Australia Day observed on the actual date
AustraliaDay1988 = Holiday(
    "Australia Day",
    month=1,
    day=26,
    start_date=Timestamp("1988-01-01"),
    end_date=Timestamp("1988-12-31"),
)
# ASX did not close for Australia Day in 1993 since
# States observed different dates prior to 1994
AustraliaDayPost88Pre93 = Holiday(
    "Australia Day",
    month=1,
    day=26,
    start_date=Timestamp("1989-01-01"),
    end_date=Timestamp("1992-12-31"),
    offset=DateOffset(weekday=MO(1)),
)

# Anzac Day was observed on Monday when it fell on a Sunday in
# 2010 but that does not appear to have been the case previously.
# ANZAC Day observance was a special case in 2010
AnzacDayNonMondayized = anzac_day(end_date="2010")
AnzacDay2010 = anzac_day(
    observance=sunday_to_monday, start_date="2010", end_date="2011"
)
AnzacDay = anzac_day(start_date="2011")

# When Easter Monday and Anzac Day coincided in 2011, Easter Tuesday was
# also observed as a public holiday. Note that this isn't defined as a
# rule, because it will happen next in 2095 (and then in  2163), and
# there isn't a great way to tell how this will be handled at that point.
EasterTuesday2011AdHoc = Timestamp("2011-04-26")

QueensBirthday = Holiday(
    "Queen's Birthday",
    month=6,
    day=1,
    offset=[DateOffset(weekday=MO(2))],
)

LastTradingDayBeforeChristmas = Holiday(
    "Last Trading Day Before Christmas",
    month=12,
    day=24,
    start_date="2010",
    observance=previous_friday,
)
Christmas = christmas()
WeekendChristmas = weekend_christmas()
BoxingDay = boxing_day()
WeekendBoxingDay = weekend_boxing_day()

LastTradingDayOfCalendarYear = Holiday(
    "Last Trading Day Of Calendar Year",
    month=12,
    day=31,
    start_date="2010",
    observance=previous_friday,
)

# additional ad-hoc holidays
NYEMonday1984AdHoc = Timestamp("1984-12-31")
NYEMonday1990AdHoc = Timestamp("1990-12-31")
Bicentennial1988 = Timestamp("1988-01-25")
Y2KTesting = Timestamp("1999-12-31")


class XASXExchangeCalendar(ExchangeCalendar):
    """
    Calendar for the Australian Securities Exchange in Sydney.

    Open Time: 10:00 AM, Australian Eastern Time
    Close Time: 4:00 PM, Australian Eastern Time

    Regularly-Observed Holidays:
      - New Year's Day
      - Australia Day
      - Good Friday
      - Easter Monday
      - Anzac Day
      - Queen's Birthday
      - Christmas Day
      - Boxing Day

    Early Closes:
      - Last trading day before Christmas
      - Last trading day of the calendar year
    """

    regular_early_close = time(14, 10)

    name = "XASX"

    tz = timezone("Australia/Sydney")

    open_times = ((None, time(10)),)  # Zipline compatability (10,1); else (10,0)

    close_times = ((None, time(16)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                AustraliaDay,
                AustraliaDayPre88,
                AustraliaDay1988,
                AustraliaDayPost88Pre93,
                GoodFriday,
                EasterMonday,
                AnzacDayNonMondayized,
                AnzacDay2010,
                AnzacDay,
                QueensBirthday,
                Christmas,
                WeekendChristmas,
                BoxingDay,
                WeekendBoxingDay,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            EasterTuesday2011AdHoc,
            NYEMonday1984AdHoc,
            NYEMonday1990AdHoc,
            Bicentennial1988,
            Y2KTesting,
        ]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        LastTradingDayBeforeChristmas,
                        LastTradingDayOfCalendarYear,
                    ]
                ),
            ),
        ]



#
# Copyright 2019 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas import Timestamp
from pandas.tseries.holiday import Easter, GoodFriday, Holiday, sunday_to_monday
from pandas.tseries.offsets import Day
from pytz import timezone

from .common_holidays import new_years_day
from .exchange_calendar import HolidayCalendar, ExchangeCalendar


class XJSEExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Johannesburg Stock Exchange (XJSE).

    Open Time: 09:00, SAST
    Close Time: 17:00, SAST

    Regularly-Observed Holidays:
    - New Year's Day (Jan 1)
    - Human Rights Day (Mar 31)
    - Good Friday (Friday before Easter)
    - Family Day (Monday after Easter)
    - Freedom Day (Apr 27)
    - Workers' Day (a.k.a. Labour Day, May 1)
    - Youth Day (Jun 16)
    - National Women's Day (Aug 19)
    - Heritage Day (Sep 24)
    - Day of Reconciliation (Dec 19)
    - Christmas (Dec 25)
    - Day of Goodwill (Dec 26)

    If a holiday falls on a Saturday, it is not observed.
    If a holiday falls on a Sunday, it is observed on Monday.

    If two holiday observations occur on the same trading day, the
    following trading day may or may not be declared a holiday: if so,
    it should be added to ``adhoc_holidays``.
    """

    name = "XJSE"

    tz = timezone("Africa/Johannesburg")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 00)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                new_years_day(
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Human Rights Day",
                    month=3,
                    day=21,
                    observance=sunday_to_monday,
                ),
                GoodFriday,
                Holiday(
                    "Family Day",
                    month=1,
                    day=1,
                    offset=[Easter(), Day(1)],
                ),
                Holiday(
                    "Freedom Day",
                    month=4,
                    day=27,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Workers' Day",
                    month=5,
                    day=1,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Youth Day",
                    month=6,
                    day=16,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "National Women's Day",
                    month=8,
                    day=9,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Heritage Day",
                    month=9,
                    day=24,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Day of Reconciliation",
                    month=12,
                    day=16,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Christmas",
                    month=12,
                    day=25,
                    observance=sunday_to_monday,
                ),
                Holiday(
                    "Day of Goodwill",
                    month=12,
                    day=26,
                    observance=sunday_to_monday,
                ),
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            Timestamp(date)
            for date in [
                # Election holidays
                "2004-04-14",
                "2006-03-01",
                "2009-04-22",
                "2011-05-18",
                "2014-05-07",
                "2016-08-03",
                "2019-05-08",
                # In 2008, Human Rights Day fell on the same day as Good
                # Friday (March 21), so the "Bridge Public Holiday" was
                # observed on May 2.
                "2008-05-02",
                # In 2011 and 2016, Christmas fell on Sunday, which would
                # ordinarily cause it to be observed on Monday; but Day of
                # Goodwill fell on Monday, which resulted in only 11
                # scheduled holidays for the year instead of the usual 12.
                # In these years, the 27th was declared a bonus holiday: it
                # may or may not be repeated in 2022 and onward.
                "2011-12-27",
                "2016-12-27",
            ]
        ]



#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import time

from pandas import Timestamp
from pandas.tseries.holiday import EasterMonday, GoodFriday, Holiday, previous_workday
from pytz import timezone

from .common_holidays import (
    boxing_day,
    christmas,
    christmas_eve,
    european_labour_day,
    new_years_day,
    new_years_eve,
    whit_monday,
)
from .exchange_calendar import HolidayCalendar, ExchangeCalendar

# Regular Holidays
# ----------------
NewYearsDay = new_years_day()

EuropeanLabourDay = european_labour_day()

# Whit Monday observed in 2007, before it became regularly observed
# starting in 2015.
WhitMonday2007AdHoc = Timestamp("2007-05-28")

# Whit Monday and the Day of German Unity have been observed regularly, but in 2022 regular trading took place instead.
#  It's unclear if it will be observed in 2023.
WhitMondayUntil2022 = whit_monday(start_date="2015-01-01", end_date="2022-01-01")

DayOfGermanUnityUntil2022 = Holiday(
    "Day of German Unity", month=10, day=3, start_date="2014-01-01", end_date="2022-01-01"
)

# Reformation Day was a German national holiday in 2017.
ReformationDay500thAnniversaryAdHoc = Timestamp("2017-10-31")

ChristmasEve = christmas_eve()

Christmas = christmas()

BoxingDay = boxing_day()

NewYearsEve = new_years_eve()

# Early Closes
# ------------
# The last weekday before Dec 31 is an early close starting in 2010.
LastWorkingDay = Holiday(
    "Last Working Day of Year Early Close",
    month=12,
    day=31,
    start_date="2010-01-01",
    observance=previous_workday,
)


class XFRAExchangeCalendar(ExchangeCalendar):
    """
    Exchange calendar for the Frankfurt Stock Exchange (XFRA).

    Open Time: 9:00 AM, CET
    Close Time: 5:30 PM, CET

    Regularly-Observed Holidays:
    - New Years Day
    - Good Friday
    - Easter Monday
    - Whit Monday
    - Labour Day
    - Day of German Unity
    - Christmas Eve
    - Christmas Day
    - Boxing Day

    Early Closes:
    - Last working day before Dec. 31
    """

    # TODO: verify the early close time
    regular_early_close = time(12, 30)

    name = "XFRA"

    tz = timezone("CET")

    open_times = ((None, time(9)),)

    close_times = ((None, time(17, 30)),)

    @property
    def regular_holidays(self):
        return HolidayCalendar(
            [
                NewYearsDay,
                GoodFriday,
                EasterMonday,
                EuropeanLabourDay,
                WhitMondayUntil2022,
                DayOfGermanUnityUntil2022,
                ChristmasEve,
                Christmas,
                BoxingDay,
                NewYearsEve,
            ]
        )

    @property
    def adhoc_holidays(self):
        return [
            WhitMonday2007AdHoc,
            ReformationDay500thAnniversaryAdHoc,
        ]

    @property
    def special_closes(self):
        return [
            (
                self.regular_early_close,
                HolidayCalendar(
                    [
                        LastWorkingDay,
                    ]
                ),
            )
        ]



# https://github.com/pandas-dev/pandas/blob/master/pandas/tseries/holiday.py

import warnings

from pandas.tseries.holiday import (
    Holiday as PandasHoliday,
    AbstractHolidayCalendar as PandasAbstractHolidayCalendar,
)
from pandas.tseries.offsets import BaseOffset
from pandas import (
    DatetimeIndex,
    Series,
    Timestamp,
)
from pandas.errors import PerformanceWarning

from .offsets import _is_normalized


class Holiday(PandasHoliday):
    """
    This extension allows to have both offset and observance while the original
    Holiday does not. The order of application is offset => observance.
    """

    def __init__(
        self,
        name,
        year=None,
        month=None,
        day=None,
        offset=None,
        observance=None,
        start_date=None,
        end_date=None,
        days_of_week=None,
        tz=None,
    ):
        super().__init__(
            name,
            year,
            month,
            day,
            None,
            None,
            start_date,
            end_date,
            days_of_week,
        )
        self.offset = offset
        self.observance = observance
        self.tz = tz
        if self.start_date is not None:
            if self.tz is None and self.start_date.tz is not None:
                self.tz = start_date.tz
            if self.start_date.tz is None and self.tz is not None:
                self.start_date = self.start_date.tz_localize(self.tz)
            assert self.tz == self.start_date.tz
        if self.end_date is not None:
            if self.tz is None and self.end_date.tz is not None:
                self.tz = start_date.tz
            if self.end_date.tz is None and self.tz is not None:
                self.end_date = self.end_date.tz_localize(self.tz)
            assert self.tz == self.end_date.tz
        if self.start_date is not None and self.end_date is not None:
            self.start_date.tz == self.end_date.tz

    def __repr__(self) -> str:
        info = ""
        if self.year is not None:
            info += f"year={self.year}, "
        info += f"month={self.month}, day={self.day}"

        if self.offset is not None:
            info += f", offset={self.offset}"

        if self.observance is not None:
            info += f", observance={self.observance}"

        repr = f"{self.__class__.__name__}: {self.name} ({info})"
        return repr

    def dates(self, start_date, end_date, return_name=False):
        start_date = Timestamp(start_date)
        end_date = Timestamp(end_date)
        assert start_date.tz == self.tz
        assert _is_normalized(start_date)
        assert end_date.tz == self.tz
        assert _is_normalized(end_date)
        return super().dates(start_date, end_date, return_name=return_name)

    def _apply_offset(self, dates):
        if self.offset is not None:
            if not isinstance(self.offset, list):
                offsets = [self.offset]
            else:
                offsets = self.offset
            for offset in offsets:
                # If we are adding a non-vectorized value
                # ignore the PerformanceWarnings:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", PerformanceWarning)
                    dates += offset
        return dates

    def _apply_observance(self, dates):
        if self.observance is not None:
            if not isinstance(self.observance, list):
                observances = [self.observance]
            else:
                observances = self.observance
            for observance in observances:
                if isinstance(observance, BaseOffset):
                    offset = observance

                    def f(d):
                        return d + offset

                    observance = f
                dates = dates.map(observance)
        return dates

    def _apply_rule(self, dates, apply_offset=True, apply_observance=True):
        if apply_offset:
            dates = self._apply_offset(dates)
        if apply_observance:
            dates = self._apply_observance(dates)
        return dates


def combine(pre_holidays):
    combined = Series(index=DatetimeIndex([]), dtype=object)
    for holidays in pre_holidays:
        combined = combined.combine_first(holidays)
    return combined


class AbstractHolidayCalendar(PandasAbstractHolidayCalendar):
    """
    This extension allows to have overlaps between calculated holidays while
    the original AbstractHolidayCalendar does not.
    """

    def holidays(self, start=None, end=None, return_name=False):
        """
        Returns a curve with holidays between start_date and end_date
        Parameters
        ----------
        start : starting date, datetime-like, optional
        end : ending date, datetime-like, optional
        return_name : bool, optional
            If True, return a series that has dates and holiday names.
            False will only return a DatetimeIndex of dates.
        Returns
        -------
            DatetimeIndex of holidays
        """
        if self.rules is None:
            raise Exception(
                f"Holiday Calendar {self.name} does not have any rules specified"
            )

        if start is None:
            start = AbstractHolidayCalendar.start_date

        if end is None:
            end = AbstractHolidayCalendar.end_date

        start = Timestamp(start)
        end = Timestamp(end)

        # If we don't have a cache or the dates are outside the prior cache, we
        # get them again
        if self._cache is None or start < self._cache[0] or end > self._cache[1]:
            pre_holidays = [
                rule.dates(start, end, return_name=True) for rule in self.rules
            ]
            if pre_holidays:
                # This line's behavior is changed to use custom combine()
                # function instead of the original concat() function
                holidays = combine(pre_holidays)
            else:
                holidays = Series(index=DatetimeIndex([]), dtype=object)

            self._cache = (start, end, holidays.sort_index())

        holidays = self._cache[2]
        holidays = holidays[start:end]

        if return_name:
            return holidays
        else:
            return holidays.index
